require 'rails_helper'

module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')

  MUTATE_PARAMS = {
    epci: 'Integer',
    max_date: "Date",
    visit_conducted_date: "Date",
    status: "String"
  }

  VALID_OPERATORS = %w[
      $in
      $nin
      $ne
      $gt
      $gte
      $lt
      $lte
      $eq
      $regex
      $all
    ].freeze

  ACCEPTABLE_PARAMS = %w[
    housing_request_id
    housing_offer_id
    epci
    max_date
    applicated_at
    effort_rate
    status_updated_at
    visit_conducted_date
  ].freeze

  describe Api::Dmo::HousingApplicationsController, :type => :api do
    routes { Hermes::Engine.routes }
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper

    let!(:responsible) {create(:user, roles: ['Administrator'])}
    let!(:user) {create(:user)}
    let!(:applicant) { create(:applicant, user: user) }

    let!(:housing_offer) {create(:housing_offer)}
    let!(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s)}

    context 'with scoring context' do
      let!(:admin) {create(:administrator)}

      let!(:threshold1) do
        create(
          :threshold,
          type: 'PLA LM',
          thresholds_map: {
            'Paris et communes limitrophes': {
              'thresholds_list': [
                13_050,
                21_272,
                27_883,
                30_521,
                36_307,
                40_859
              ],
              'another_person': 4552
            },
            'IDF hors Paris et Communes limitrophes': {
              'thresholds_list': [
                13_050,
                21_272,
                25_569,
                28_075,
                33_238,
                37_401
              ],
              'another_person': 4166
            },
            'Autres régions': {
              'thresholds_list': [
                11_342,
                16_525,
                19_872,
                22_111,
                25_870,
                29_155
              ],
              'another_person': 3252
            }
          }
        )
      end

      let!(:user1) {create(:user)}
      let!(:applicant1) do
        create(
            :applicant,
            sne_profile_creation_date: '2016-09-24',
            user_id: user1.id,
            amount_year_minus_two: 9787.24,
            monthly_incomes: {
              'Salaire': 1045.23,
              'APL': 321.62
            }
        )
      end
      let!(:request1) do
        create(
            :housing_request,
            is_active: true,
            main_applicant: applicant1,
            seeked_districts: ['34507'],
            seeked_departments: ['34']
        )
      end

      let!(:user2) {create(:user)}
      let!(:applicant2) do
        create(
            :applicant,
            sne_profile_creation_date: '2017-09-24',
            user_id: user2.id,
            amount_year_minus_two: 9787.24,
            monthly_incomes: {
                'Salaire': 1045.23,
                'APL': 321.62
            }
        )
      end
      let!(:request2) do
        create(
            :housing_request,
            is_active: true,
            main_applicant: applicant2,
            seeked_districts: ['34507'],
            seeked_departments: ['34']
        )
      end

      let(:group1) {create(:group, user_ids: [user1.id])}

      let!(:offer1) do
        create(
            :housing_offer,
            rent_amount: 234.32,
            rental_charges: 42.15,
            parking_charges: 12.71,
            surface: 29,
            threshold_type: 'PLA LM',
            offer_status: 'LOGEMENT PUBLIE',
            department: '34',
            threshold_subtype: 'IDF hors Paris et Communes limitrophes',
            readers: ["g_#{group1.id}"]
        )
      end

      let!(:offer2) do
        create(
          :housing_offer,
          rent_amount: 434.32,
          rental_charges: 53.12,
          parking_charges: 22.43,
          surface: 34,
          threshold_type: 'PLA LM',
          offer_status: 'LOGEMENT PUBLIE',
          department: '34',
          threshold_subtype: 'Autres régions',
          readers: ["g_#{group1.id}"]
        )
      end

      describe 'retrieving applciation scoring data' do
        before :each do
          perform_enqueued_jobs do
            enqueued_jobs.each do |job|
              job[:job].perform_now(*job[:args])
            end
          end
          authenticated_header(request, responsible)
          post 'create', {
            data: {
              attributes: {
                housing_offer_id: offer1.id.to_s,
                housing_request_id: request1.id.to_s
              }
            }
          }
        end

        it 'should has valid scoring datas' do
          perform_enqueued_jobs do
            enqueued_jobs.each do |job|
              job[:job].perform_now(*job[:args])
            end
          end
          attributes = json['data']['attributes']
          expect(request1.reload.housing_offer_ids).to include(offer1.id.to_s)
          expect(
            Hermes::Score.find_by(
              housing_offer: offer1,
              applicant: request1.main_applicant
            ).has_applicated
          ).to be true
          expect(Hermes::HousingApplication.unscoped.count).to be 1
          expect(
            Hermes::HousingApplication.find_by(
              housing_request: request1,
              housing_offer: offer1
            ).id.to_s
          ).to eq json['data']['id']
          expect(
            Hermes::HousingApplication.where(
              housing_request: request1,
              housing_offer: offer1
            ).count
          ).to be 1
         # expect(request1.reload.last_application_date).to eq DateTime.parse(attributes['created_at'])
          # expect(
          #   Hermes::HousingApplication.find_by(
          #     housing_request: request1,
          #     housing_offer: offer1
          #   ).score_id&.to_s
          # ).to eq attributes['score_id']
          # expect(
          #   Hermes::HousingApplication.unscoped.find_by(
          #     housing_request: request1,
          #     housing_offer: offer1
          #   ).score.score_with_position
          # ).to be 20

          expect(attributes['score_id']).to be_present
          expect(
            attributes.slice(
              *%w[
                profile_epci_score
                eligible_position
                applicated_position
              ]
            ).values
          ).to all be_a Integer
        end

        describe 'GET application for actualized scoring' do
          before :each do
            perform_enqueued_jobs do
            enqueued_jobs.each do |job|
              job[:job].perform_now(*job[:args])
            end
          end
            authenticated_header(request, responsible)
            get 'index'
          end

          it 'has valid score_with_position' do
            expect(json['data'].length).to be 1
            expect(
              json['data'].first['attributes']['score_with_position']
            ).to be 20
          end
        end
      end
    end

    describe 'POST /housing_applications' do
      it 'should has no housing_applciaiton' do
        expect(housing_request.applications_nb).to be 0
      end

      describe 'Create housing_application with minimal valid datas' do
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
            data: {
              attributes: {
                housing_offer_id: housing_offer.id.to_s,
                housing_request_id: housing_request.id.to_s
              }
            }
          }
        end
        it 'returns status code 201' do
          expect(response).to have_http_status(201)
        end

        it 'should update housing_request applciations_nb' do
          expect(housing_request.reload.applications_nb).to be 1
        end
      end
    end

    describe 'GET /housing_applications' do
      describe 'GET /housing_applications' do
        let!(:housing_application1) do
          create(
            :housing_application,
            housing_offer_id: housing_offer.id.to_s,
            housing_request_id: housing_request.id.to_s
          )
        end

        let!(:housing_application3) do
          create(
            :housing_application,
            housing_offer_id: housing_offer.id.to_s,
            housing_request_id: housing_request.id.to_s
          )
        end

        let!(:housing_application2) do
          create(
            :housing_application,
            housing_offer_id: housing_offer.id.to_s,
            housing_request_id: housing_request.id.to_s
          )
        end

        before :each do
          authenticated_header(request, responsible)
          get 'index'
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
          expect(json['data'].length).to eq(3)
          expect(json['meta']['pagination']).to be_present
          expect(json['meta']['pagination']['total_objects']).to eq(3)
        end
      end
    end

    describe 'GET /housing_applications/:id' do
      describe 'verif it work' do
        let!(:housing_application1) {create(:housing_application, housing_offer_id: housing_offer.id.to_s,
                                            housing_request_id: housing_request.id.to_s)}
        before :each do
          authenticated_header(request, responsible)
          get 'show', {id: housing_application1.id}
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
        end
      end
      describe 'verif it work not work with a non authorized user' do
        let!(:housing_application1) {create(:housing_application, housing_offer_id: housing_offer.id.to_s,
                                            housing_request_id: housing_request.id.to_s)}

        before :each do
          authenticated_header(request, user)
          get 'show', {id: housing_application1.id}
        end
        it 'returns status code 404' do
          expect(response).to have_http_status(404)
        end
      end
    end

    describe 'PATCH /housing_applications/:id' do
      let!(:user2) {create(:user)}
        let!(:housing_application) {create(:housing_application, user_id: user2.id.to_s, housing_offer_id: housing_offer.id.to_s,
                                           housing_request_id: housing_request.id.to_s,   status: "EN ATTENTE",)}
        before :each do
          authenticated_header(request, responsible)
      
          patch 'update', {
            id: housing_application.id.to_s, 
            data: {
              attributes: {
                status: "ACCEPTEE EN CAL"
              }
            }
          }
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
          expect(json["data"]["attributes"]["responsible"]["id"]).to eq(user2.id.to_s)
        end
    end

    describe 'DELETE /housing_applications/:id' do
      let!(:housing_application) {create(:housing_application, user_id: responsible.id.to_s, housing_offer_id: housing_offer.id.to_s,
                                         housing_request_id: housing_request.id.to_s)}
      before :each do
        authenticated_header(request, responsible)

        delete 'destroy', {
            id: housing_application.id.to_s
        }
      end
      it 'returns status code 404' do
        expect(response).to have_http_status(204)
      end
    end
  end

  require_relative '../../swagger_helper'

  describe Hermes::Api::Dmo::HousingApplicationsController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
    TAGS_API_DMO_HOUSING_APPLICATIONS= 'API_DMO_HOUSING_APPLICATIONS'
    DESCRIPTION_STATUS_HTML= '<b>Correspondance entre satus de la housing application et les Codes utilisés par sopra</b><br> '\
    '<i>la donnée à renseigner doit correspondre exactement a cette nomenclature afin de pouvoir declencher des flux vers sopra</i>'\
    '<br><table style="border-collapse:collapse;border-spacing:0" class="tg"><tr><th style="font-family:Arial, sans-serif;font-size:14px;font-weight:normal;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">code sopra</th><th style="font-family:Arial, sans-serif;font-size:14px;font-weight:normal;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">status</th></tr><tr><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">INSCAL</td><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">"Inscription CAL"</td></tr><tr><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">DESCAL</td><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">"Attribution lot avec rang"</td></tr><tr><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">DESCAL01</td><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">"Attribution lot"</td></tr><tr><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">DESCAL02</td><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">"Attribution sous condition"</td></tr><tr><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">DESCAL03</td><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">"Refus attribution"</td></tr><tr><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">REFLOG</td><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">"Refus lot demandeur"</td></tr><tr><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">VISPRO</td><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">"Visite proposée"</td></tr><tr><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">VISEFF</td><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">"Visite effectuée"</td></tr><tr><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">RADATT</td><td style="font-family:Arial, sans-serif;font-size:14px;padding:10px 5px;border-style:solid;border-width:1px;overflow:hidden;word-break:normal;border-color:black;text-align:left;vertical-align:top">"Radiation attribution"</td></tr></table>'

    include JwtAuthHelper
    extend HermesSwaggerJsonParams
    include SwaggerJsonAttributes
    let!(:responsible) {create(:user, first_name: 'Martin', last_name: 'Durand', email: 'martin.durand@gmail.com', roles: ['Administrator'])}
    let!(:user) {create(:user, first_name: 'Jackie', last_name: 'Mike', email: 'jackie.mike@gmail.com')}
    let!(:applicant) { create(:applicant, user: responsible) }
    let!(:housing_offer) {create(:housing_offer)}
    let!(:housing_request) {create(:housing_request, main_applicant_id: applicant.id.to_s)}

    context 'as admin' do
      let!(:admin) {create(:administrator)}

      path '/api/dmo/housing_applications' do
        post 'Create housing_application' do
          tags TAGS_API_DMO_HOUSING_APPLICATIONS
          produces 'application/json'
          consumes 'application/json'
          security [Bearer: {}]
          description DESCRIPTION_STATUS_HTML

          json_params={}

          json_params[:housing_offer_id]={type: :string}
          json_params[:housing_request_id]={type: :string}
          MUTATE_PARAMS.each do |key, val|
            json_params[key]= {type: nice_type(val)}
          end
          parameter name: :params, in: :body, schema: jsonify_params(json_params)

          response '201', 'housing_request created' do
            let(:Authorization) {"Bearer #{token_for(responsible)}"}
            let(:params) do
              jsonapi_attributes(
                  {housing_offer_id: housing_offer.id.to_s,
                  housing_request_id:housing_request.id.to_s}
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'

            after do |_example|
              json = JSON.parse(response.body)

              expect(json['data']).to be_a Hash
              expect(json['data']['id']).to_not be nil
              expect(json['data']['type']).to eq 'housing_applications'
            end
          end
        end
      end

      path '/api/dmo/housing_applications' do
        get 'Retrieves all housing_applications' do
          tags TAGS_API_DMO_HOUSING_APPLICATIONS
          produces 'application/json'

          parameter name: 'ids[]', in: :query, type: :array, items: { type: :string }, collectionFormat: :multi, required: false
          ACCEPTABLE_PARAMS.each do |acceptable_param|
            type = :string
            parameter name: acceptable_param.to_s, in: :query, type: type, required: false
          end
        # VALID_OPERATORS.each do |valid_operator|
          #  parameter( name: valid_operator.to_s, in: :query,type: :object, required: false)
        # end

          security [Bearer: {}]
          paginator

          response '200', 'housing_applications found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:housing_applications) { create_list(:housing_application, 5,housing_offer_id: housing_offer.id.to_s,
                                                      housing_request_id: housing_request.id.to_s
            ) }

            include_context 'with integration test'
            include_context 'valid jsonapi collection' do
              let(:expected_collection) { housing_applications }
            end
          end
        end
      end

      path '/api/dmo/housing_applications/{id}' do
        get 'Retrieves a single housing_application by id' do
          tags TAGS_API_DMO_HOUSING_APPLICATIONS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]

          response '200', 'housing_application found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let(:housing_application) { create(:housing_application,housing_offer_id: housing_offer.id.to_s,
                                               housing_request_id: housing_request.id.to_s
            ) }
            let(:id) { housing_application.id }

            include_context 'with integration test'
            include_context 'valid jsonapi ressource' do
              let(:expected_ressource) { housing_application }
            end
          end
        end
      end

      path '/api/dmo/housing_applications/{id}' do
        put 'edit housing_application' do
          tags TAGS_API_DMO_HOUSING_APPLICATIONS
          produces 'application/json'
          consumes 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]
          json_params = {}
          json_params[:housing_offer_id] = { type: :string }
          json_params[:housing_request_id] = { type: :string }
          MUTATE_PARAMS.each do |key, val|
            json_params[key]= {type: nice_type(val)}
          end
          description DESCRIPTION_STATUS_HTML

          parameter name: :params, in: :body, schema: jsonify_params(json_params)
          #jsonapi_params( **json_params )

          response '200', 'housing_application updated' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:housing_application) { create(:housing_application,housing_offer_id: housing_offer.id.to_s,
                                               housing_request_id: housing_request.id.to_s
            ) }
            let(:id) {housing_application.id.to_s}
            let(:params) do
              jsonapi_attributes(
                {
                  status: "ACCEPTEE EN CAL"
                }
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'

            after do |_example|
              json = JSON.parse(response.body)

              expect(json['data']).to be_a Hash
              expect(json['data']['id']).to_not be nil
              expect(json['data']['type']).to eq 'housing_applications'
            end
          end
        end
      end

      path '/api/dmo/housing_applications/{id}' do
        delete 'Delete a single housing_application by id' do
          tags TAGS_API_DMO_HOUSING_APPLICATIONS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]

          response '204', 'housing_application deleted' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let(:id) { create(:housing_application, user_id: admin.id,housing_offer_id: housing_offer.id.to_s,
                housing_request_id: housing_request.id.to_s
            ).id }
            run_test!
          end
        end
      end
    end
  end
end
